import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:math';
import 'package:hotech_flutter_utilities/Constants.dart';
import 'package:hotech_flutter_utilities/model/tools/CacheDir.dart';
import 'package:hotech_flutter_utilities/service/LoginHelper.dart';
import 'package:hotech_flutter_utilities/service/ServiceFactory.dart';
import 'package:flutter/material.dart';
import 'package:http/http.dart';
import 'package:intl/intl.dart';
import 'package:path_provider/path_provider.dart';
import 'package:pic/pic.dart';
import 'package:sprintf/sprintf.dart';
import 'package:url_launcher/url_launcher.dart';
import 'package:mime_type/mime_type.dart';
import 'package:http_parser/http_parser.dart';

typedef GlobalCallBack<T> = void Function(T value);

class HFUtils {
  ///
  /// taken from flutter/material/constant class
  ///
  static const double toolbarHeight = 56.0;

  static Size getScreenSize(BuildContext context) {
    return MediaQuery.of(context).size;
  }

  static double getScreenHeight(BuildContext context) {
    return getScreenSize(context).height;
  }

  static double getScreenWidth(BuildContext context) {
    return getScreenSize(context).width;
  }

  static double getStatusBarHeight(BuildContext context) {
    return MediaQuery.of(context).padding.top;
  }

  ///
  /// IphoneX and later notched Iphones have a "Home Bar" at the bottom, this causes extra padding to be present
  /// at the bottom than normal, this method returns that extra padding.
  ///
  static double getBottomNavBarExtraHeight(BuildContext context) {
    return MediaQuery.of(context).padding.bottom;
  }

  //not used YET
  static bool isLargeAspectRatio(BuildContext context) {
    return getScreenWidth(context) / getScreenHeight(context) > 0.8;
  }

  ///
  /// height of screen excluding statusBar, Bottom Navigation Bar, toolbar(Appbar) height, and any extra padding caused
  /// by the notched Iphones' bottom home bar.
  ///
  static double getPureScreenHeight(BuildContext context,
      {bool hasToolbar = true, bool hasNavBar = false, bool bottomPadding = true}) {
    return getScreenHeight(context) -
        getStatusBarHeight(context) -
        (bottomPadding ? getBottomNavBarExtraHeight(context) : 0.0) -
        (hasNavBar ? toolbarHeight : 0.0) -
        (hasToolbar ? toolbarHeight : 0.0);
  }

  static void navigate(BuildContext context, StatefulWidget aWidget) {
    Navigator.push(
      context,
      MaterialPageRoute(builder: (context) {
        return aWidget;
      }),
    );
  }

  static String getViewField(String field) {
    return Consts.VIEW_FIELD + field;
  }

  static bool isDEBUG() {
    return !bool.fromEnvironment("dart.vm.product");
  }

  static void launchGoogleMapsUrl(double lat, double lng) async {
    final url = sprintf(Consts.GOOGLE_MAP_URL, [lat.toString(), lng.toString()]);
    if (await canLaunch(url)) {
      await launch(url);
    } else {
      print("Could not launch URL");
    }
  }

  static void logPrint(Object object) async {
    int defaultPrintLength = 1020;
    if (object == null || object.toString().length <= defaultPrintLength) {
      print(object);
    } else {
      String log = object.toString();
      int start = 0;
      int endIndex = defaultPrintLength;
      int logLength = log.length;
      int tmpLogLength = log.length;
      while (endIndex < logLength) {
        print(log.substring(start, endIndex));
        endIndex += defaultPrintLength;
        start += defaultPrintLength;
        tmpLogLength -= defaultPrintLength;
      }
      if (tmpLogLength > 0) {
        print(log.substring(start, logLength));
      }
    }
  }

  ///
  /// [getRowList] uses [getNRowList] to return a row containing a number of items specified in [itemsPerRow], useful
  /// when needing a specific number of items per line or per slider, if the item count is not a multiple of [itemsPerRow],
  /// the remaining items are added to an extra row ([getRowList(list of 8 items, itemsPerRow: 5)] ->  row 1: 5 items, row2: 3 items)
  ///

  static List<Row> getRowList(List<Widget> widgetList, {int itemsPerRow = 2}) {
    List<Row> listofRows = new List<Row>();
    int modulo = widgetList.length % itemsPerRow;
    if (modulo == 0) {
      listofRows.addAll(getNRowList(widgetList, widgetList.length, itemsPerRow));
    } else {
      listofRows.addAll(getNRowList(widgetList, widgetList.length - modulo, itemsPerRow));
      listofRows.add(new Row(
          mainAxisAlignment: MainAxisAlignment.center,
          children: widgetList.getRange(widgetList.length - modulo, widgetList.length).toList()));
    }
    return listofRows;
  }

  ///
  ///gets rows for a multiple of [n] number of items, THIS METHOD IS ONLY MEANT TO BE CALLED BY [getRowList], USE THAT INSTEAD
  ///
  static List<Row> getNRowList(List<Widget> widgetList, int length, int n) {
    List<Row> listofNRows = new List<Row>();
    for (int i = 0; i < length; i += n) {
      Row rowToAdd = new Row(
        mainAxisAlignment: MainAxisAlignment.spaceEvenly,
        children: widgetList.getRange(i, i + n).toList(),
      );
      listofNRows.add(rowToAdd);
    }
    return listofNRows;
  }

  static String capitalizeOnlyFirst(String word) {
    if (word.length != null) {
      return word[0].toUpperCase() + word.substring(1, word.length).toLowerCase();
    }
    return '';
  }

  static NumberFormat numberFormat;

  static bool checkBright(File imageToCheck) {
    int r = 0;
    int g = 0;
    int b = 0;
    bool isBright = true;
    if (imageToCheck != null) {
      if (imageToCheck.existsSync()) {
        Pic picc = decodeImage(imageToCheck.readAsBytesSync());
        List<int> imagepixels = picc.getBytes();
        List<int> topPixels = imagepixels.getRange(0, ((imagepixels.length / 4) ~/ 3 * 4).toInt()).toList();
        for (int pixel = 0; pixel < topPixels.length; pixel += 4) {
          r += topPixels[pixel];
          g += topPixels[pixel + 1];

          b += topPixels[pixel + 2];
        }
        r = r ~/ (topPixels.length / 4);
        g = g ~/ (topPixels.length / 4);
        b = b ~/ (topPixels.length / 4);

        int brightness = sqrt(0.241 * pow(r, 2) + 0.691 * pow(g, 2) + 0.068 * pow(b, 2)).toInt();
        isBright = (brightness > 130);
      }
    }
    return isBright;
  }
}

///
/// Caching related utilities
///

class CacheUtils {
  static Map<String, CacheDir> cacheDirMap = {};
  static List<CacheDir> cacheDirList = [];
  static String appDocsDir;
  static File cacheDirFile;
  static bool isappDocsDirLoaded = false;
  static bool isCacheDirMapLoaded = false;

  ///
  /// Retrieves the directory where files are stored
  ///
  static Future<Directory> getCacheDir({bool isRafile = false}) async {
    String dir;
    if (!isappDocsDirLoaded) //only start the retrieval process if the directory hasn't been retrieved already
    {
      Directory directory = await getApplicationDocumentsDirectory();
      appDocsDir = directory.path;
      dir = appDocsDir;
      isappDocsDirLoaded = true;
    } else {
      dir = appDocsDir;
    }
    //adds the full path of the stored files to the documents directory
    String path = "$dir/${DBConst.FDN_FILES}/${Consts.PARAM_NETWORKCACHE}${isRafile ? "/${DBConst.TBL_RAFILE}" : ""}";
    Directory directory = Directory(path);
    directory.createSync(recursive: true); //create the directory if it doesn't exist
    return directory;
  }

  ///
  /// Gets the index file where a list of all cached files is kept
  ///

  static Future<File> getCacheDirFile() async {
    if (cacheDirFile == null) {
      Directory directory = await getCacheDir();
      cacheDirFile = File("${directory.path}/dir.json");
      return cacheDirFile;
    } else {
      return cacheDirFile;
    }
  }

  ///
  /// create a map in memory with where all the files stored (faster than reading the index file every time)
  ///
  static Future getCacheDirMap() async {
    if (!isCacheDirMapLoaded) {
      File file = await getCacheDirFile();
      List<dynamic> jsonList = file.existsSync() ? json.decode(file.readAsStringSync()) : [];
      jsonList.forEach((jsonItem) {
        cacheDirList.add(CacheDir.fromJson(jsonItem));
      });
      cacheDirList.forEach((cacheDir) {
        cacheDirMap[cacheDir.key] = cacheDir;
      });
      isCacheDirMapLoaded = true;
    }
  }

  ///
  /// Writes the CacheDirMap to the index file
  ///

  static Future setCacheDir() async {
    File file = await getCacheDirFile();
    List<Map<String, dynamic>> jsonList = [];
    cacheDirMap.values.forEach((cacheDir) {
      jsonList.add(cacheDir.toJson());
    });
    file.writeAsStringSync(json.encode(jsonList));
  }

  ///
  ///deletes all records in cache and their files
  ///

  static Future invalidateCache() async {
    File file = await getCacheDirFile();
    file.writeAsStringSync("{}");
    Directory dir = await getCacheDir();
    dir.deleteSync();
  }

  ///
  /// deletes one record from the cache and the file associated with that record
  ///

  static Future invalidateCacheEntry(String key) async {
    String path = cacheDirMap[key].filePath;
    cacheDirMap.remove(key);
    File(path).deleteSync();
    await setCacheDir();
  }

  ///
  /// Returns the same name if no other file with that name exists in the same folder or a
  /// unique name if a fle or more exist(s), (image.jpg -> image(0).jpg -> image(1).jpg)
  ///

  static File getUniqueName(String path, String name, {int num = -1}) {
    List<String> nameParts = name.split('.');
    String extension = nameParts.removeLast();
    String newName = "";
    nameParts.forEach((part) {
      newName += part + '.';
    });
    newName = newName.substring(0, newName.length - 1);
    newName += '(' + num.toString() + ')' + '.' + extension;
    File file = (num == -1) ? File(path + name) : File(path + newName);
    num++;
    if (file.existsSync())
      return getUniqueName(path, name, num: num);
    else {
      file.writeAsBytesSync([], flush: true);
      return file;
    }
  }

  static HttpClient httpClient = HttpClient();
}

///
/// Rafile (files from the service) related utilities
///

class RafileUtils {
  static Future<MultipartRequest> getMultipartRequest(
    Uri uri,
    List<File> files, {
    bool printLog = true,
  }) async {
    MultipartRequest multipartRequest = MultipartRequest(Consts.STR_POST, uri);
    multipartRequest.headers.addAll(await LoginHelper.getHeaders());

    for (File file in files) {
      MultipartFile multipartFile = await MultipartFile.fromPath(
        DBConst.FDN_FILE,
        "${file.path}",
        filename: getFileName(file),
        contentType: MediaType.parse(mime(getFileName(file))),
      );
      multipartRequest.files.add(multipartFile);
    }

    if (printLog) {
      String responseStr = "\n";
      multipartRequest.files.forEach((multiPartFile) {
        responseStr += "\t${multiPartFile.field}: ${multiPartFile.filename}\n";
        responseStr += "\t${Consts.CONTENT_TYPE}: ${multiPartFile.contentType}\n";
        responseStr += "\t${Consts.CONTENT_LENGTH}: ${multiPartFile.length}";
      });

      ServiceFactory.printRequestLog(
        uri.toString(),
        HttpMethod.POST,
        multipartRequest.headers,
        responseStr,
      );
    }

    return multipartRequest;
  }

  static String getFileName(File file) {
    return file != null ? file.path.split("/").last : "";
  }

  static Future<File> getFile(int masterid, String name, bool getNew) async {
    Directory cacheDir = await CacheUtils.getCacheDir(isRafile: true);
    Directory directory = Directory("${cacheDir.path}/$masterid/");
    directory.createSync(recursive: true);
    return getNew ? CacheUtils.getUniqueName(directory.path, name) : File("${directory.path}$name");
  }
}
