import 'dart:async';
import 'dart:convert';

import 'package:hotech_flutter_utilities/Constants.dart';
import 'package:hotech_flutter_utilities/model/SingleDataResult.dart';
import 'package:hotech_flutter_utilities/model/ListDataResult.dart';
import 'package:hotech_flutter_utilities/service/LoginHelper.dart';
import 'package:hotech_flutter_utilities/service/ServiceResponse.dart';
import 'package:hotech_flutter_utilities/service/SyncRequest.dart';
import 'package:hotech_flutter_utilities/service/SyncResponse.dart';
import 'package:http/http.dart';
import 'package:preferences/preferences.dart';

///
/// http method enum [HttpMethod.GET] and [HttpMethod.POST]
///
enum HttpMethod { GET, POST }

abstract class ServiceFactory<T> {
  ///
  /// this returns the utilities Url (base url) unless otherwise specified.
  ///
  String getBaseUrl() => PrefService.getString(Consts.PREF_KEY_SYNC_URL);

  ///
  /// abstract method to get the entity (table) name.
  ///
  String getTableName();

  ///
  /// a method that handles consuming restFull services
  ///
  Future<Response> callService(
    HttpMethod method,
    String moduleUrl,
    Map<String, String> headers, {
    requestBody,
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Duration timeout = const Duration(seconds: 30),
  }) async {
    // get url
    String url = getBaseUrl() + (getTableName() ?? "") + moduleUrl + buildParamQuery(params, queries);

    try {
      Future<Response> future;
      Response response;

      // print the request sent along with method name, token, and date to the console.
      printRequestLog(url, method, headers, requestBody);

      // consume the service.
      if (method == HttpMethod.GET) {
        future = get(url, headers: headers);
      } else if (method == HttpMethod.POST) {
        future = post(url, body: requestBody, headers: headers);
      }

      if (future != null) {
        // get the response.
        response = timeout != null ? await future.timeout(timeout) : await future;

        // print the response in a formatted way.
        if (response != null) {
          printResponseLog(method, response, url);
          return response;
        }
      }
    } catch (exception) {
      print(exception);
    }

    return null;
  }

  ///
  /// print the request sent
  ///
  static void printRequestLog(
    String url,
    HttpMethod method,
    headers,
    requestBody,
  ) {
    String httpMethod = method == HttpMethod.GET ? Consts.STR_GET : method == HttpMethod.POST ? Consts.STR_POST : "";
    String header = headers != null ? headers.toString() + "\n" : "";
    String responseStr = "\n$httpMethod --> $url\n$header${Consts.STR_HOST}: ${Uri.parse(url).host}\n";

    if (method == HttpMethod.POST && requestBody != null) {
      responseStr = "$responseStr${Consts.STR_BODY_REQUEST}: $requestBody\n";
    }

    print(responseStr);
  }

  ///
  /// print the response received from the service along with status code, url, and response.
  /// takes the response as a [String]
  ///
  void printStrResponseLog(
    HttpMethod method,
    String response,
    int statusCode,
    String url,
  ) {
    String responseStr =
        "\n$statusCode <-- $url\n${Consts.STR_DATE}: ${DateTime.now().toIso8601String()}\n${Consts.STR_RESPONSE} $response\n\n";

    print(responseStr);
  }

  ///
  /// print the response received from the service along with status code, url, and response.
  ///
  void printResponseLog(
    HttpMethod method,
    Response response,
    String url,
  ) {
    printStrResponseLog(method, response.body, response.statusCode, url);
  }

  ///
  /// @return the [query] param and the other params to be appended to the services
  ///
  /// @example {entity}/view/list?query=id:15,roomno:15D&allhotels=true&limit=50
  ///
  static String buildParamQuery(Map<String, dynamic> params, Map<String, dynamic> queries) {
    String ret = "";
    String paramsStr = buildParam(params);
    String queriesStr = buildQuery(queries);
    if (paramsStr != "") {
      paramsStr = "?$paramsStr";
      ret = paramsStr;
      if (queriesStr != "") {
        ret = "$ret&" + queriesStr;
      }
    } else if (queriesStr != "") {
      queriesStr = "?$queriesStr";
      ret = queriesStr;
    }
    return ret;
  }

  ///
  /// @return the list of params to be appended to the services
  ///
  /// @example params are [allhotels] and [limit] and the result is
  /// {entity}/view/list?allhotels=true&limit=50
  ///
  static String buildParam(Map<String, dynamic> params) {
    String ret = "";
    if (params == null) {
      return ret;
    }
    params.forEach((key, value) {
      ret = "${ret != "" ? ret + "&" : ret}$key=$value";
    });
    return ret != "" ? "$ret" : ret;
  }

  ///
  /// @return the [query] param to be appended to the services
  ///
  /// @example {entity}/view/list?query=id:15,roomno:15D
  ///
  static String buildQuery(Map<String, dynamic> params) {
    String ret = "";
    if (params == null) {
      return ret;
    }
    params.forEach((key, value) {
      ret = "${ret != "" ? ret + "," : ret}$key:$value";
    });
    return ret != "" ? "${DBConst.FDN_QUERY}=$ret" : ret;
  }

  ///
  /// get the response from {entity}/view/list services.
  ///
  /// @return as [Response]
  ///
  Future<Response> getViewListResponse({
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    Response response = await callService(
      HttpMethod.GET,
      Consts.PATH_VIEW_LIST,
      headers ?? await LoginHelper.getHeaders(),
      params: params,
      queries: queries,
    );

    return response;
  }

  ///
  /// get the response from {entity}/view/list services.
  /// @return as [List] of objects.
  ///
  Future<List<T>> getViewList({
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    Response response = await getViewListResponse(params: params, queries: queries, headers: headers);

    return response != null && ResponseType.isSuccessful(response.statusCode)
        ? ListDataResult<T>.fromJson(json.decode(response.body)).data
        : null;
  }

  ///
  /// sync services
  ///
  /// @example {entity}/sync
  ///
  Future<SingleDataResult<SyncResponse<T>>> syncRequest(SyncRequest request) {
    return null;
  }

  ///
  /// ins services
  ///
  /// @example {entity}/ins
  ///
  Future<T> insService(
    request, {
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
  }) async {
    Response response = await callService(
      HttpMethod.POST,
      Consts.PATH_INS,
      await LoginHelper.getPostHeaders(),
      requestBody: request,
      params: params,
      queries: queries,
    );
    return response != null ? SingleDataResult<T>.fromJson(json.decode(response.body)).data : null;
  }

  Future<Response> insServiceResponse(
    request, {
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    Response response = await callService(
      HttpMethod.POST,
      Consts.PATH_INS,
      headers ?? await LoginHelper.getPostHeaders(),
      requestBody: request,
      params: params,
      queries: queries,
    );

    return response;
  }

  ///
  /// @return the [Response] object from {entity}/view/list/get/{field} service.
  ///
  Future<Response> getListGetResponse(
    String field,
    List values, {
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    var response = await callService(
      HttpMethod.POST,
      Consts.PATH_VIEW_LIST + Consts.PATH_GET + "/" + field,
      headers ?? await LoginHelper.getPostHeaders(),
      requestBody: json.encode(values),
      params: params,
      queries: queries,
    );
    return response;
  }

  ///
  /// @return the [List] of records from {entity}/view/list/get service.
  ///
  Future<List<T>> getListGet(
    String field,
    List values, {
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    var response = await getListGetResponse(field, values, params: params, queries: queries, headers: headers);
    return response != null ? ListDataResult<T>.fromJson(json.decode(response.body)).data : null;
  }

  ///
  /// @return the [Response] object of {entity}/def service
  ///
  Future<Response> getDefResponse({
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    var response = await callService(
      HttpMethod.GET,
      Consts.PATH_DEF,
      headers ?? await LoginHelper.getHeaders(),
      params: params,
      queries: queries,
    );
    return response;
  }

  ///
  /// @return the [T] object of {entity}/def service
  ///
  Future<T> getDef({
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, String> headers,
  }) async {
    var response = await getDefResponse(params: params, queries: queries, headers: headers);
    return response != null ? SingleDataResult<T>.fromJson(json.decode(response.body)).data : null;
  }
}
