import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:hotech_flutter_utilities/Constants.dart';
import 'package:hotech_flutter_utilities/model/SingleDataResult.dart';
import 'package:hotech_flutter_utilities/model/ListDataResult.dart';
import 'package:hotech_flutter_utilities/model/tools/CacheDir.dart';
import 'package:hotech_flutter_utilities/model/tools/rafile.dart';
import 'package:hotech_flutter_utilities/service/LoginHelper.dart';
import 'package:hotech_flutter_utilities/service/ServiceResponse.dart';
import 'package:hotech_flutter_utilities/service/ServiceFactory.dart';
import 'package:hotech_flutter_utilities/service/SyncRequest.dart';
import 'package:hotech_flutter_utilities/service/SyncResponse.dart';
import 'package:hotech_flutter_utilities/HFUtils.dart';
import 'package:http/http.dart';

class RafileService extends ServiceFactory<Rafile> {
  static RafileService instance;

  static RafileService getInstance() {
    return instance != null ? instance : RafileService();
  }

  @override
  String getTableName() {
    return DBConst.TBL_RAFILE;
  }

  @override
  Future<SingleDataResult<SyncResponse<Rafile>>> syncRequest(SyncRequest request) async {
    return null;
  }

  Future<File> download(
    Rafile raFile,
    int masterId, {
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, dynamic> headers,
  }) async {
    await CacheUtils.getCacheDirMap();
    String cacheKey = "${raFile.id}";
    File file;
    if (!CacheUtils.cacheDirMap.containsKey(cacheKey) ||
        CacheUtils.cacheDirMap[cacheKey].updateTime != raFile.updatetime) {
      file = await RafileUtils.getFile(masterId, raFile.filename, true);
      file = await downloadFile(file, raFile.updatetime, cacheKey, params: params, queries: queries, headers: headers);
    } else {
      file = await RafileUtils.getFile(masterId, raFile.filename, false);
    }
    return file;
  }

  Future<File> downloadFile(
    File file,
    String updateTime,
    String cacheKey, {
    Map<String, dynamic> params,
    Map<String, dynamic> queries,
    Map<String, dynamic> headers,
  }) async {
    var service = await callService(
      HttpMethod.GET,
      Consts.PATH_DOWNLOAD,
      headers ?? await LoginHelper.getHeaders(),
      params: params,
      queries: queries,
    );
    try {
      //print("Downloading --> ${service.request}");
      file.writeAsBytesSync(service.bodyBytes);
      CacheUtils.cacheDirMap[cacheKey] = CacheDir(key: cacheKey, filePath: file.path, updateTime: updateTime);
    } catch (e) {
      print("exc during downloading ${file.path} with this message $e");
      deleteFileIfNotSuccessful(service, file, cacheKey);
      return null;
    }
    await deleteFileIfNotSuccessful(service, file, cacheKey);
    return file;
  }

  Future deleteFileIfNotSuccessful(Response response, File file, String cacheKey) async {
    if (!ResponseType.isSuccessful(response.statusCode)) {
      await CacheUtils.invalidateCacheEntry(cacheKey);
      print("Could not retrieve file, response code is: ${response.statusCode}");
    } else {
      print("File Downloaded Successfully, response code is: ${response.statusCode}");
    }
  }

  Future<Map<String, dynamic>> getUploadResponse(String aServiceName, int masterId, List<File> files) async {
    try {
      String safeMasterIdParam = masterId != 0 ? "?${DBConst.FDN_MASTERID}=$masterId" : "";

      Uri url = Uri.parse(
        getBaseUrl() + getTableName() + aServiceName + safeMasterIdParam,
      );

      MultipartRequest request = await RafileUtils.getMultipartRequest(url, files);
      StreamedResponse streamedResponse = await request.send();
      if (streamedResponse != null &&
          streamedResponse.stream != null &&
          ResponseType.isSuccessful(streamedResponse.statusCode)) {
        var responseDecoded = json.decode(
          await streamedResponse.stream.bytesToString(),
        );
        printStrResponseLog(
          HttpMethod.POST,
          json.encode(responseDecoded),
          streamedResponse.statusCode,
          url.toString(),
        );
        return responseDecoded;
      }
    } on Exception catch (e) {
      print("exc in upload $e}");
    }
    return null;
  }

  Future<Rafile> upload(int masterId, File files) async {
    try {
      Map<String, dynamic> responseDecoded = await getUploadResponse(
        Consts.PATH_UPLOAD,
        masterId,
        [files],
      );
      return responseDecoded != null ? SingleDataResult<Rafile>.fromJson(responseDecoded).data : null;
    } on Exception catch (e) {
      print("exc in upload $e}");
    }
    return null;
  }

  Future<List<Rafile>> multiUpload(int masterId, List<File> files) async {
    try {
      Map<String, dynamic> responseDecoded = await getUploadResponse(Consts.PATH_MULTI_UPLOAD, masterId, files);
      return responseDecoded != null ? ListDataResult<Rafile>.fromJson(responseDecoded).data : null;
    } on Exception catch (e) {
      print("exc in multiUpload $e}");
    }
    return null;
  }
}
